#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <netdb.h>
#include <stdlib.h>
#include <iostream>
#include <cstring>
#include <sys/poll.h>
#include <fcntl.h>
#include <unordered_map>

#include "server_util.h"
#include "threadpool.h"
#include "server_socket.h"

#include "FileParser.h" // File reading

static const int num_threads = 16;

using namespace std;

// This is the function that threads are dispatched into
// in order to process new client connections.
static void HttpServer_ThrFn(ThreadPool::Task *t);

// given a program name, it prints
// information about its usage.
void Usage(char *progname);

int main(int argc, char **argv) {
  // We need 1 pollfd for the listening socket, num_tread pollfd for client connections,
  // and num_threads pollfd for the workers.
  // We are using a fixed size number of clients == 16 == num_threads == num workers
  // events[0].fd will always be the listening_socket fd
  // then for each i > 0 and i is even, events[i].fd = client_fd and events[i + 1].fd = worker_fd (read end of pipe)
  // and we have a mapping from worker_fd(read end of the pipe) -> Work that each worker has to do
  struct pollfd events[1 + (num_threads * 2)];
  int num_fds = 1;
  int poll_timeout = -1; // -1 is same as infinite timeout
  bool resize_events_arr = false;
  bool new_client_added = false;
  unordered_map<int, Work*> worker_to_client;  // read end of a worker -> work to do map

  // Check for valid number of arguments
  if (argc != 3)
    Usage(argv[0]);

  // Parse the port number, and check
  // for failure
  unsigned short port = 0;
  if (sscanf(argv[2], "%hu", &port) != 1)
    Usage(argv[0]);

  ServerSocket listen_socket(port, argv[1]);

  // Try to bind the socket and start listening for connections
  if (!listen_socket.BindAndListen()) {
    cerr << "Socket bind and listen failed." << endl;
    return EXIT_FAILURE;
  }

  // Threadpool of workers
  ThreadPool tp(num_threads);

  // initialize pollfd structs for event polling
  memset(events, 0, sizeof(events));

  events[0].fd = listen_socket.GetSocketFd();
  events[0].events = POLLIN;

  // Loop and poll for events

  while (1) {
    // poll for events. Current time out is -1, which means the server
    // will wait forever until some events can be polled

    int poll_res = poll(events, num_fds, poll_timeout);

    if (poll_res < 0) {
      perror("   poll failed");
      return EXIT_FAILURE;
    } else if (poll_res == 0) {
      // happens if we set timeout other than -1
      perror("   poll timed out");
      return EXIT_FAILURE;
    }
    
    // we have at least one event
    int curr_num_fs = num_fds;
    for (int i = 0; i < curr_num_fs; ++i) {
      // Loop through to find the descriptors that returned
      // POLLIN and determine whether it's the listening
      // or the active connection.
      if (events[i].revents == 0)
        continue;

      // If revents is not POLLIN, it's an unexpected result,
      // ignore, and repoll for other events
      if (events[i].revents != POLLIN) {
        cerr << " events at i = " << i << ", reveents = " << events[i].revents << "\n";
        break;
      }

      if (events[i].fd == listen_socket.GetSocketFd()) {
        if (new_client_added) {
          // if we already accepted a new client connection, we need to set it up first
          continue;
        }
        cout << "received new client \n";
        // we have a new client connection
        int new_client_fd;

        // Accept client connection
        // if we encounter error with this event the go to next event
        if (!listen_socket.Accept(&new_client_fd)) {
          continue;
        }

        events[num_fds].fd = new_client_fd;
        events[num_fds].events = POLLIN;
        num_fds++;
        new_client_added = true;

        cout << "new client  added\n";

      } else if (events[i].fd != -1) {
        if (i % 2 != 0) {
          // odd indices store the client fds
          
          // the client wrote the file path to us
          string req_file_path;
          int bytes_read = WrappedRead(events[i].fd, &req_file_path);
          if (bytes_read > PATH_MAX) {
            // file path too large
            close(events[i].fd);
            events[i].fd = -1;
            new_client_added = false;
            resize_events_arr = true;
            continue;
          }

          // read file path from client and do path validation
          string base_dir(".");

          if (!IsPathSafe(base_dir, req_file_path)) {
            perror(" Path is not safe");
            close(events[i].fd);
            events[i].fd = -1;
            new_client_added = false;
            resize_events_arr = true;

            continue;
          }
          // all good, dispatch thread and setup pollfd 

          // setup pipe for worker thread
          int pipe_fds[2];
          if (pipe2(pipe_fds, O_NONBLOCK) == -1) {
            perror("  pipe failed");
            close(events[i].fd);
            events[i].fd = -1;
            new_client_added = false;
            resize_events_arr = true;

            continue;
          }
          
          // Initialize Work struct for worker thread
          Work *thread_work = new Work;
          thread_work->filepath = req_file_path;
          // pipe_fds[1] is write end of pipe
          thread_work->pipe_fd = pipe_fds[1];

          // add pollfd 
          // pipe_fds[0] is the read end of the pipe
          events[num_fds].fd = pipe_fds[0];
          events[num_fds].events = POLLIN;
          num_fds++;
          new_client_added = false;

          // dispatch thread
          HttpServerTask *worker_task = new HttpServerTask(HttpServer_ThrFn);
          worker_task-> work = thread_work;

          worker_to_client[pipe_fds[0]] = thread_work;

          tp.Dispatch(worker_task);
        } else {
          // this is a worker
          // this is a worker thread done with the work
          Work *thread_work = worker_to_client[events[i].fd];
          if (thread_work == nullptr) {
            break;
          }

          int client_fd = events[i - 1].fd;
          // return data to client
          // Note: we do not need locks to read the data from the shared buffer,
          // this is because we read the buffer ONLY when the worker thread notifies us.
          // The worker thread notifies the main thread iff the worker is done
          // writing to the shared buffer.
          int written_bytes = WrappedWrite(client_fd, (unsigned char*)(thread_work->buffer.c_str()), thread_work->buffer.size());
          if ((size_t)written_bytes != thread_work->buffer.size()) {
            cerr << "client_fd: " << client_fd;
            cerr << ", actual written: " << written_bytes << ", expected: " << thread_work->buffer.size() << "\n";
          }

          // clean up resource: pipes, client socket, work struct
          close(client_fd);
          close(events[i].fd);
          close(thread_work->pipe_fd);

          delete thread_work;
          
          events[i].fd = -1; // read end of worker pipe
          events[i - 1].fd = -1; // client connection
          // We deleted two events, so set resize_events_arr to true, 
          // so that we can resize our events array
          resize_events_arr = true;
        }

      }
    }

    // at least 1 event has been deleted, need to resize
    if (resize_events_arr) {
      resize_events_arr = false;
      for (int i = 0; i < num_fds; i++) {
        if (events[i].fd == -1) {
          for(int j = i; j < num_fds; j++) {
            events[j].fd = events[j+1].fd;
          }
          i--;
          num_fds--;
        }
      }
    }
  }
  

  return EXIT_SUCCESS;
}

static void HttpServer_ThrFn(ThreadPool::Task *t)
{
  // Input: Shared struct with path, buffer, write end of pipe, and client_fd(only need for main thread)
  // 1. A worker function that receives the name of a file (absolute file path),
  HttpServerTask *hst = static_cast<HttpServerTask *>(t);
  string fullfile = hst->work->filepath;

  // 2. reads the file content into the address space of the process,
  int file_len = 0;
  char *file = ReadFileToString(fullfile.c_str(), &file_len);
  if (file == NULL) {
    cout << "file pointer is null\n";
  }

  // 3. Append file contents to the shared buffer
  // Note: we do not need any locks because the main thread will only read
  // from this shared buffer iff the worker thread write to the pipe.
  hst->work->buffer.clear();
  hst->work->buffer.append(file, file_len);

  // 4. Tell main thread that task is done, main thread needs to close the connection
  char check = 'c';
  int write_res = 0;
  while (write_res != 1) {
    write_res = WrappedWrite(hst->work->pipe_fd, (unsigned char *)&check, 1);
  }

  // free the file, because the contents have already been moved to the shared buffer
  free(file);
}

void Usage(char *progname) {
  std::cerr << "usage: " << progname << " network_address port_num" << std::endl;
  exit(EXIT_FAILURE);
}
