#include <stdio.h>       // for snprintf()
#include <unistd.h>      // for close(), fcntl()
#include <sys/types.h>   // for socket(), getaddrinfo(), etc.
#include <sys/socket.h>  // for socket(), getaddrinfo(), etc.
#include <arpa/inet.h>   // for inet_ntop()
#include <netdb.h>       // for getaddrinfo()
#include <errno.h>       // for errno, used by strerror()
#include <string.h>      // for memset, strerror()
#include <iostream>      // for std::cerr, etc.
#include <fcntl.h>

#include "./server_socket.h"

extern "C" {
  #include "./verify550.h" // For asserts

}

using namespace std;

ServerSocket::ServerSocket(uint16_t port, char *hostname) {
  port_ = port;
  listen_sock_fd_ = -1;
  host_name_ = hostname;
  sock_family_ = AF_INET; // only for IPV4 addresses
}

ServerSocket::~ServerSocket() {
  // Close the listening socket if it's not zero.  The rest of this
  // class will make sure to zero out the socket if it is closed
  // elsewhere.
  if (listen_sock_fd_ != -1)
    close(listen_sock_fd_);
  listen_sock_fd_ = -1;
}

bool ServerSocket::BindAndListen() {
  // Use "getaddrinfo," "socket," "bind," and "listen" to
  // create a listening socket on port port_.  Return the
  // listening socket through the output parameter "listen_fd".

  struct addrinfo hints;
  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_family = sock_family_;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE;
  hints.ai_flags |= AI_V4MAPPED;
  hints.ai_protocol = IPPROTO_TCP;
  hints.ai_canonname = nullptr;
  hints.ai_addr = nullptr;
  hints.ai_next = nullptr;

  std::string port = std::to_string(port_);
  struct addrinfo *result;
  int res = getaddrinfo(host_name_, port.c_str(), &hints, &result);
  // check that getaddrinfo was successful
  if (res != 0)
    return false;

  for (struct addrinfo *rp = result; rp != nullptr; rp = rp->ai_next) {
    listen_sock_fd_ = socket(rp->ai_family,
                             rp->ai_socktype,
                             rp->ai_protocol);
    // if we fail to create a socket, then continue to
    // the next returned result, and try again.
    if (listen_sock_fd_ == -1) {
      listen_sock_fd_ = 0;
      continue;
    }

    // configure the socket, tell the tcp stack to make the
    // port we bind available again as soon as we exit.
    int optval = 1;
    if (setsockopt(listen_sock_fd_, SOL_SOCKET, SO_REUSEADDR,
               &optval, sizeof(optval)) == -1) {

      perror(" setsock failed");
    }
    // if we successfully binded to a socket, then exit the loop
    if (bind(listen_sock_fd_, rp->ai_addr, rp->ai_addrlen) == 0) {
      break;
    }
    perror(" bind failed");

    // the bind failed, so we close the socket, and try again with
    // the next result
    close(listen_sock_fd_);
    listen_sock_fd_ = -1;
  }
  freeaddrinfo(result);

  // check if binding was a success,
  // if not return false.
  if (listen_sock_fd_ == -1)
    return false;


  // success, tell the OS that we want this to be a listening socket.
  if (listen(listen_sock_fd_, SOMAXCONN) != 0) {
    close(listen_sock_fd_);
    return false;
  }

  int fcntl_status = fcntl(listen_sock_fd_, F_SETFL, fcntl(listen_sock_fd_, F_GETFL, 0) | O_NONBLOCK);

  if (fcntl_status == -1) {
    perror("fcntl failed");
    // handle the error.  By the way, I've never seen fcntl fail in this way
    return false;
  }

  return true;
}

bool ServerSocket::Accept(int *accepted_fd) {
  // Accept a new connection on the listening socket listen_sock_fd_.
  // (Block until a new connection arrives.)  Return the newly accepted
  // socket through the output parameter.
  struct sockaddr_storage caddr;
  socklen_t caddr_len = sizeof(caddr);

  // accept client connection, close the listening socket if it fails
  int client_fd = accept4(listen_sock_fd_,
                         reinterpret_cast<struct sockaddr *>(&caddr),
                         &caddr_len, O_NONBLOCK);
  if (client_fd < 0) {
    if (errno != EWOULDBLOCK && errno != EINTR) {
      close(listen_sock_fd_);
      perror("   accept failed");
      exit(EXIT_FAILURE);
    }
    return false;
  }
  *accepted_fd = client_fd;

  return true;
}

int ServerSocket::GetSocketFd() {
  return listen_sock_fd_;
}

