#ifndef SERVER_SOCKET_H_
#define SERVER_SOCKET_H_

#include <netdb.h>       // for AF_UNSPEC, AF_INET, AF_INET6
#include <stdint.h>      // for uint16_t, etc.
#include <sys/types.h>   // for AF_UNSPEC, AF_INET, AF_INET6
#include <sys/socket.h>  // for AF_UNSPEC, AF_INET, AF_INET6
#include <string>        // for std::string

// A ServerSocket class abstracts away the messy details of creating a
// TCP listening socket at a specific port and on a (hopefully)
// externally visible IP address.  As well, a ServerSocket helps
// customers accept incoming client connections on the listening
// socket.
class ServerSocket {
 public:
  // This constructor creates a new ServerSocket object and associates
  // it with the provided port number.  The constructor doesn't create
  // a socket yet; it just memorizes the given port.
  explicit ServerSocket(uint16_t port, char *hostname);

  // The destructor closes the listening socket if it is open.
  virtual ~ServerSocket();

  // This function causes the ServerSocket to attempt to create a
  // listening socket and to bind it to the given port number on
  // whatever IP address the host OS recommends for us.  The caller
  // provides:
  //
  // On failure this function returns false.  On success, it returns
  // true, sets listen_sock_fd_ to be the file descriptor for the
  // listening socket.
  bool BindAndListen();

  // This function causes the ServerSocket to attempt to accept
  // an incoming connection from a client.  On failure, returns false.
  // On success, it returns true, and also returns (via output
  // parameters) the following:
  //
  // - accepted_fd: the file descriptor for the new client connection.
  //   The customer is responsible for close()'ing this socket when it
  //   is done with it.
  //
  bool Accept(int *accepted_fd);

  int GetSocketFd();

 private:
  uint16_t port_;
  int listen_sock_fd_;
  int sock_family_;  // either AF_INET or AF_INET6 for ipv4 or ipv6/v4
  char *host_name_;
};

#endif  // SERVER_SOCKET_H_
