// Feature test macro enabling strdup (c.f., Linux Programming Interface p. 63)
#define _XOPEN_SOURCE 600

#include "./FileParser.h"

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Constants and declarations of internal helper functions

#define ASCII_UPPER_BOUND 0x7F


///////////////////////////////////////////////////////////////////////////////
// Publically-exported functions

char *ReadFileToString(const char *filename, int *size) {
    struct stat filestat;
    char *buf;
    int fd;
    ssize_t numread;
    size_t left_to_read;

    // STEP 1.
    // Use the stat system call to fetch a "struct stat" that describes
    // properties of the file. ("man 2 stat"). You can assume we're on a 64-bit
    // system, with a 64-bit off_t field.
    if (stat(filename, &filestat) == -1) {  // stat failed
        perror(" stat failed");
        return NULL;
    }


    // STEP 2.
    // Make sure this is a "regular file" and not a directory or something else
    // (use the S_ISREG macro described in "man 2 stat").
    if (!S_ISREG(filestat.st_mode)) {  // not a regular file
        perror(" not reg file");
        return NULL;
    }


    // STEP 3.
    // Attempt to open the file for reading (see also "man 2 open").
    fd = open(filename, O_RDONLY);
    if (fd == -1) {
        perror(" failed to open");
        close(fd);
        return NULL;
    }


    // STEP 4.
    // Allocate space for the file, plus 1 extra byte to
    // '\0'-terminate the string.
    *size = filestat.st_size + 1;
    buf = (char *) malloc(*size * sizeof(char));
    if (buf == NULL) {  // malloc failed

        perror(" malloc failed");
        close(fd);
        return NULL;
    }

    // STEP 5.
    // Read in the file contents using the read() system call (see also
    // "man 2 read"), being sure to handle the case that read() returns -1 and
    // errno is either EAGAIN or EINTR.  Also, note that it is not an error for
    // read to return fewer bytes than what you asked for; you'll need to put
    // read() inside a while loop, looping until you've read to the end of file
    // or a non-recoverable error.  Read the man page for read() carefully, in
    // particular what the return values -1 and 0 imply.
    left_to_read = filestat.st_size;
    while (left_to_read > 0) {
        numread = read(fd, buf, left_to_read);
        if (numread == -1) {
            // either interrupted or an actual error
            if (errno != EINTR || errno != EAGAIN) {
                // Some kind of error.

                perror(" read failed");
                free(buf);
                close(fd);
                return NULL;
            }
        } else {
            left_to_read -= numread;
        }
    }

    // Great, we're done!  We hit the end of the file and we read
    // filestat.st_size - left_to_read bytes. Close the file descriptor returned
    // by open() and return through the "size" output parameter how many bytes
    // we read.
    close(fd);
    *size = filestat.st_size - left_to_read;

    // Null-terminate the string.
    buf[*size] = '\0';
    return buf;
}