import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import dslabs.atmostonce.AMOApplication;
import dslabs.atmostonce.AMOCommand;
import dslabs.atmostonce.AMOResult;
import dslabs.framework.Address;
import dslabs.framework.Application;
import dslabs.framework.Command;
import dslabs.framework.Node;
import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import lombok.Data;
import lombok.EqualsAndHashCode;
import lombok.ToString;


@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
public class PaxosServer extends Node {
    private final Address[] servers;

    private BallotNumber ballot;                              // Our ballot number
    private BallotNumber highestBallotSeen;                   // Highest ballot that we saw
    private Map<Integer, BallotValuePair> proposals;          // Used by leader to send P2A to acceptors
    private Map<Integer, BallotValuePair> accepted;           // Used by acceptors in P2A and also P1B
    private Map<Integer, BallotValuePair> log;                // The log
    private boolean isLeader;                                 // Are we the leader?
    private AMOApplication app;                               // The kv store
    private Set<PaxosRequest> requestInProcess;               // Request in progress so we dont double propose
    private boolean leaderRecentPing;                         // If this is true we think the leader is alive

    // For proposal phase
    private int slotOut;                                      // First non-executed proposal slot (+1 last executed slot)
    private int slotIn;                                       // First empty slot (+1 last proposed slot)
    private Multimap<Integer, Address> slotsToAcceptors;      // Used by leader to decide majority for each slot after P2B

    // For leader election
    private Set<Address> voters;                              // Yes votes for leader election
    private Map<Integer, BallotValuePair> p1bReplies;         // Accepted values for each slot by acceptors that voted yes

    // For garbage collection
    private Map<Address, Integer> replicasLastExecuted;       // Map from replicas to last executed slot
    private int lastGarbageCollectedSlot;                     // Slot of last garbage collected slot in our local

    /* -------------------------------------------------------------------------
        Construction and Initialization
       -----------------------------------------------------------------------*/
    public PaxosServer(Address address, Address[] servers, Application app) {
        super(address);
        this.servers = servers;

        // NOTE SLOT NUMBERS START FROM 1
        this.ballot = new BallotNumber(1, this.address());
        this.log = new HashMap<>();
        this.highestBallotSeen = new BallotNumber(0, this.address());
        this.proposals = new HashMap<>();
        this.accepted = new HashMap<>();
        this.slotsToAcceptors = HashMultimap.create();
        this.slotOut = 1;
        this.slotIn = 1;
        this.replicasLastExecuted = new HashMap<>();
        this.isLeader = false;
        this.lastGarbageCollectedSlot = 0;
        this.app = new AMOApplication(app);
        this.requestInProcess = new HashSet<>();
        this.leaderRecentPing = false;

        this.voters = new HashSet<>();
        this.p1bReplies = new HashMap<>();

        // On start servers[0] is the leader hard coded
        if (address().equals(servers[0])) {
            this.isLeader = true;
            ballot.increaseBallot();
        } else {
            highestBallotSeen = new BallotNumber(2, servers[0]);
        }
    }


    @Override
    public void init() {
        set(new HeartBeatTimer(), HeartBeatTimer.HEARTBEAT_RETRY_MILLIS);
        set(new HeartBeatCheckTimer(highestBallotSeen), HeartBeatCheckTimer.HEARTBEAT_CHECK_RETRY_MILLIS);
    }

    /* -------------------------------------------------------------------------
        Interface Methods

        Be sure to implement the following methods correctly. The test code uses
        them to check correctness more efficiently.
       -----------------------------------------------------------------------*/

    /**
     * Return the status of a given slot in the server's local log.
     *
     * If this server has garbage-collected this slot, it should return {@link
     * PaxosLogSlotStatus#CLEARED} even if it has previously accepted or chosen
     * command for this slot. If this server has both accepted and chosen a
     * command for this slot, it should return {@link PaxosLogSlotStatus#CHOSEN}.
     *
     * Log slots are numbered starting with 1.
     *
     * @param logSlotNum
     *         the index of the log slot
     * @return the slot's status
     *
     * @see PaxosLogSlotStatus
     */
    public PaxosLogSlotStatus status(int logSlotNum) {
        // Make sure the orders of if statements is right
        // For example, Chosen should be above Accepted since chosen value
        // is also on the accepted map
        if (logSlotNum <= lastGarbageCollectedSlot) {
            return PaxosLogSlotStatus.CLEARED;
        }
        if (log.containsKey(logSlotNum)) {
            return PaxosLogSlotStatus.CHOSEN;
        }
        if (accepted.containsKey(logSlotNum)) {
            return PaxosLogSlotStatus.ACCEPTED;
        }
        return PaxosLogSlotStatus.EMPTY;
    }

    /**
     * Return the command associated with a given slot in the server's local
     * log.
     *
     * If the slot has status {@link PaxosLogSlotStatus#CLEARED} or {@link
     * PaxosLogSlotStatus#EMPTY}, this method should return {@code null}.
     * Otherwise, return the command this server has chosen or accepted,
     * according to {@link PaxosServer#status}.
     *
     * If clients wrapped commands in {@link dslabs.atmostonce.AMOCommand}, this
     * method should unwrap them before returning.
     *
     * Log slots are numbered starting with 1.
     *
     * @param logSlotNum
     *         the index of the log slot
     * @return the slot's contents or {@code null}
     *
     * @see PaxosLogSlotStatus
     */
    public Command command(int logSlotNum) {
        PaxosLogSlotStatus status = status(logSlotNum);
        if (status == PaxosLogSlotStatus.CLEARED || status == PaxosLogSlotStatus.EMPTY) {
            return null;
        }
        if (status == PaxosLogSlotStatus.CHOSEN) {
            PaxosRequest pr = log.get(logSlotNum).value();
            // No Op
            if (pr.value() == null) {
                return null;
            }
            return pr.value().command();
        }
        // Status == Accepted
        PaxosRequest pr = accepted.get(logSlotNum).value();
        // No Op
        if (pr.value() == null) {
            return null;
        }
        return pr.value().command();
    }

    /**
     * Return the index of the first non-cleared slot in the server's local log.
     * The first non-cleared slot is the first slot which has not yet been
     * garbage-collected. By default, the first non-cleared slot is 1.
     *
     * Log slots are numbered starting with 1.
     *
     * @return the index in the log
     *
     * @see PaxosLogSlotStatus
     */
    public int firstNonCleared() {
        return lastGarbageCollectedSlot + 1;
    }

    /**
     * Return the index of the last non-empty slot in the server's local log,
     * according to the defined states in {@link PaxosLogSlotStatus}. If there
     * are no non-empty slots in the log, this method should return 0.
     *
     * Log slots are numbered starting with 1.
     *
     * @return the index in the log
     *
     * @see PaxosLogSlotStatus
     */
    public int lastNonEmpty() {
        int last = slotIn;
        for (int i = last; i >= 0; i--) {
            if (status(i) != PaxosLogSlotStatus.EMPTY) {
                last = i;
                break;
            }
        }

        return last;
    }

    /* -------------------------------------------------------------------------
        Message Handlers
       -----------------------------------------------------------------------*/
    private void handlePaxosRequest(PaxosRequest m, Address sender) {
        if (!isLeader) {
            // Drop if not leader
            return;
        }
        if (app.alreadyExecuted(m.value())) {
            AMOResult r = app.execute(m.value());
            if (r != null) {
                send(new PaxosReply(r), r.clientAddress());
            }
            return;
        }
        if (requestInProcess.contains(m)) {
            // Used to make sure shat we dont use two log slot for the same
            // request
            return;
        }

        BallotValuePair bvp = new BallotValuePair(new BallotNumber(ballot.sequenceNum(), address()), m);
        proposals.put(slotIn, bvp);
        accepted.put(slotIn, bvp);
        P2A p2a = new P2A(new BallotNumber(ballot.sequenceNum(), address()), slotIn, m);
        // Increase slotIn
        slotIn++;
        // P2A
        for (Address acceptor : servers) {
            if (!acceptor.equals(address())) {
                send(p2a, acceptor);
            }
        }

        // Accept our own proposal
        slotsToAcceptors.put(p2a.slotNum(), address());

        requestInProcess.add(p2a.value());
        set(new P2ATimer(p2a), P2ATimer.P2A_RETRY_MILLIS);

        if (slotsToAcceptors.get(p2a.slotNum()).size() > servers.length/2) {
            // Majority accepted, can put into log
            log.put(p2a.slotNum(), bvp);
            executeLog();
        }
    }

    private void handleP1A(P1A p1a, Address sender) {
        // If the ballot of the sender is higher than any ballot it has seen:
        // 1. Update its highest ballot number seen
        // 2. Respond with a message (P1B) containing accepted values
        if (highestBallotSeen.compareTo(p1a.ballotNumber()) < 0) {
            highestBallotSeen = p1a.ballotNumber();

            // What if this server is the leader? Do we make this a "follower"?
            if (isLeader && p1a.ballotNumber().compareTo(ballot) > 0) {
                isLeader = false;
            }

            P1B p1b = new P1B(highestBallotSeen, accepted);
            send(p1b, sender);
            set(new HeartBeatCheckTimer(p1a.ballotNumber()), HeartBeatCheckTimer.HEARTBEAT_CHECK_RETRY_MILLIS * 2);
            leaderRecentPing = true;
        }
    }

    private void handleP1B(P1B p1b, Address sender) {
        if (isLeader) {
            return;
        }
        // If receive majority of response from acceptors with its ballot, becomes leader
        if (ballot.equals(p1b.acceptedBallot()) && ballot.equals(highestBallotSeen)) {
            voters.add(sender);
            for (Map.Entry<Integer, BallotValuePair> e : p1b.accepted().entrySet()) {
                Integer slot = e.getKey();
                BallotValuePair new_bvp = e.getValue();

                BallotValuePair cur_bvp = p1bReplies.get(slot);

                if (cur_bvp == null) {
                    p1bReplies.put(slot, new_bvp);
                } else {
                    if (new_bvp.ballotNumber().compareTo(cur_bvp.ballotNumber()) >= 0) {
                        p1bReplies.put(slot, new_bvp);
                    }
                }
            }
            if (voters.size() > servers.length/2 - 1) {
                // This server is elected as leader
                // Must update its state with accepted values from acceptors
                isLeader = true;
                for (int slot : p1bReplies.keySet()) {
                    if (status(slot) != PaxosLogSlotStatus.CHOSEN) {
                        proposals.put(slot, new BallotValuePair(new BallotNumber(ballot.sequenceNum(), address()), p1bReplies.get(slot).value()));
                        slotIn = Math.max(slotIn, slot + 1);
                        BallotValuePair bvp = proposals.get(slot);

                        P2A p2a = new P2A(bvp.ballotNumber(), slot, bvp.value());
                        for (Address acceptor : servers) {
                            if (!acceptor.equals(address())) {
                                send(p2a, acceptor);
                            }
                        }
                        accepted.put(slot, new BallotValuePair(new BallotNumber(ballot.sequenceNum(), address()), p1bReplies.get(slot).value()));
                        // Accept our own proposal
                        slotsToAcceptors.put(p2a.slotNum(), address());

                        requestInProcess.add(p2a.value());
                        set(new P2ATimer(p2a), P2ATimer.P2A_RETRY_MILLIS);
                    }
                }
            }
        }
    }

    private void handleP2A(P2A p2a, Address sender) {
        if (!highestBallotSeen.equals(p2a.ballot())) {
            // Dont care if it is not from leader
            return;
        }
        if (status(p2a.slotNum()) == PaxosLogSlotStatus.CLEARED) {
            return;
        }
        if (accepted.containsKey(p2a.slotNum())) {
            BallotValuePair bvp = accepted.get(p2a.slotNum());
            if (bvp.ballotNumber().compareTo(p2a.ballot()) <= 0) {
                // Bigger ballot or equal, replace and send accepted
                bvp = new BallotValuePair(p2a.ballot(), p2a.value());
                accepted.put(p2a.slotNum(), bvp);
            } else {
                // Don't do anything
                return;
            }
        } else {
            // If we have not accepted anything then accept it
            BallotValuePair bvp = new BallotValuePair(p2a.ballot(), p2a.value());
            accepted.put(p2a.slotNum(), bvp);
        }
        requestInProcess.add(p2a.value());
        slotIn = Math.max(slotIn, p2a.slotNum() + 1);

        P2B p2b = new P2B(p2a.ballot(), p2a.slotNum());
        send(p2b, sender);
    }

    private void handleP2B(P2B p2b, Address sender) {
        if (!isLeader) {
            // Stop if no longer leader
            return;
        }
        if (status(p2b.slotNum()) == PaxosLogSlotStatus.CLEARED) {
            return;
        }
        // check if it is still consistent with our proposal
        BallotValuePair bvp = proposals.get(p2b.slotNum());
        if (!bvp.ballotNumber().equals(p2b.ballot())) {
            // No longer in that proposal
            return;
        }

        // Keep track of who have accepted
        slotsToAcceptors.put(p2b.slotNum(), sender);
        if (slotsToAcceptors.get(p2b.slotNum()).size() > servers.length/2) {
            // Majority accepted, can put into log
            log.put(p2b.slotNum(), bvp);
            executeLog();
        }
    }

    public void handleLeaderHeartbeat(LeaderHeartbeat heartbeat, Address sender) {
        if (highestBallotSeen.compareTo(heartbeat.ballotNumber()) > 0) {
            // Not leader
            return;
        }
        boolean newLeaderSeen = false;
        if (highestBallotSeen.compareTo(heartbeat.ballotNumber()) < 0) {
            isLeader = false;
            highestBallotSeen = heartbeat.ballotNumber();
            newLeaderSeen = true;
        }
        // Replace log with the bigger log slot
        for (int slot : heartbeat.log().keySet()) {
            slotIn = Math.max(slotIn, slot + 1);
            if (!log.containsKey(slot)) {
                log.put(slot, heartbeat.log().get(slot));
            } else {
                BallotValuePair bvp = log.get(slot);
                BallotValuePair new_bvp = heartbeat.log().get(slot);
                if (new_bvp.ballotNumber().compareTo(bvp.ballotNumber()) > 0) {
                    log.put(slot, new_bvp);
                }
            }
        }
        executeLog();
        // Garbage Collection that is actually working, can use this function below as helper
        garbageCollect(heartbeat.garbageCollectUntil());

        leaderRecentPing = true;
        HeartbeatResponse hbr = new HeartbeatResponse(slotOut - 1, highestBallotSeen);
        send(hbr, sender);
        if (newLeaderSeen){
            set(new HeartBeatCheckTimer(highestBallotSeen), HeartBeatCheckTimer.HEARTBEAT_CHECK_RETRY_MILLIS * 2);
        }
    }

    public void handleHeartbeatResponse(HeartbeatResponse heartbeatResponse, Address sender) {
        if (!isLeader) {
            return;
        }
        int lastExecuted = replicasLastExecuted.getOrDefault(sender, 0);
        lastExecuted = Math.max(lastExecuted, heartbeatResponse.lastExecutedSlot());
        replicasLastExecuted.put(sender, lastExecuted);
    }

    private void garbageCollect(int garbageCollectUntil) {
        if (garbageCollectUntil <= lastGarbageCollectedSlot) {
            return;
        }
        for (int i = lastGarbageCollectedSlot; i <= garbageCollectUntil; i++) {
            log.remove(i);
            accepted.remove(i);
            slotsToAcceptors.removeAll(i);
            proposals.remove(i);
            p1bReplies.remove(i);
        }
        lastGarbageCollectedSlot = garbageCollectUntil;
    }

    /* -------------------------------------------------------------------------
        Timer Handlers
       -----------------------------------------------------------------------*/
    public void onP2ATimer(P2ATimer t) {
        // If not leader then stop doing it
        if (isLeader && status(t.p2a().slotNum()) != PaxosLogSlotStatus.CHOSEN
            && status(t.p2a().slotNum()) != PaxosLogSlotStatus.CLEARED) {
            for (Address acceptor : servers) {
                if (!acceptor.equals(address())) {
                    send(t.p2a(), acceptor);
                }
            }
            set(t, P2ATimer.P2A_RETRY_MILLIS);
        }
    }

    public void onHeartBeatTimer(HeartBeatTimer t) {
        executeLog();
        if (isLeader) {
            for (Address acceptor : servers) {
                if (!acceptor.equals(address())) {
                    LeaderHeartbeat lh = new LeaderHeartbeat(log, garbageCollectUntil(), ballot);
                    send(lh, acceptor);
                }
            }
            int lastEx = replicasLastExecuted.getOrDefault(address(), 0);
            replicasLastExecuted.put(address(), Math.max(slotOut - 1, lastEx));
            garbageCollect(garbageCollectUntil());
        }
        set(t, HeartBeatTimer.HEARTBEAT_RETRY_MILLIS);
    }

    public void onHeartBeatCheckTimer(HeartBeatCheckTimer t) {
        if (isLeader) {
            return;
        }
        if (t.ballotNumber().equals(highestBallotSeen)) {
            // Check if the leader alive or not
            if (!leaderRecentPing) {
                // Leader is dead
                // Just for randomization for contention issue
                electLeader();
                return;
            }
            leaderRecentPing = false;
            set(t, HeartBeatCheckTimer.HEARTBEAT_CHECK_RETRY_MILLIS);
        }
    }

    public void onLeaderElectionTimer(LeaderElectionTimer t) {
        if (highestBallotSeen.equals(ballot) && !isLeader) {
            for (Address acceptor : servers) {
                if (!acceptor.equals(address())) {
                    send(t.p1a(), acceptor);
                }
            }
            set(t, LeaderElectionTimer.LEADER_ELECTION_TIMER);
        }
    }

    private void electLeader() {
        // Try to elect ourself as the leader
        // Try to get elected as leader at the beginning of time
        voters.clear();
        p1bReplies.clear();
        p1bReplies.putAll(accepted);

        // Increase ballot until higher than the highest we saw before electing
        while (ballot.compareTo(highestBallotSeen) < 0) {
            ballot.increaseBallot();
        }

        P1A p1a = new P1A(new BallotNumber(ballot.sequenceNum(), address()));
        highestBallotSeen = p1a.ballotNumber();
        
        // P1A
        for (Address acceptor : servers) {
            if (!acceptor.equals(address())) {
                send(p1a, acceptor);
            }
        }
        
        set(new LeaderElectionTimer(p1a), LeaderElectionTimer.LEADER_ELECTION_TIMER);
    }

    /* -------------------------------------------------------------------------
        Utils
       -----------------------------------------------------------------------*/

    // Helper method to execute log as much as we can
    private void executeLog() {
        boolean foundEmpty = false;
        for (int i = slotOut; i < slotIn; i++) {
            PaxosLogSlotStatus status = status(i);
            if (status == PaxosLogSlotStatus.CHOSEN) {
                if (foundEmpty) {
                    continue;
                }
                BallotValuePair bvp = log.get(i);
                AMOCommand cmd = bvp.value().value();
                if (cmd != null) {
                    AMOResult r = app.execute(cmd);
                    requestInProcess.remove(bvp.value());
                    send(new PaxosReply(r), r.clientAddress());
                }
                slotOut++;
            } else {
                foundEmpty = true;
                if (status == PaxosLogSlotStatus.EMPTY) {
                    // NO-OP
                    if (isLeader) {
                        P2A p2a = new P2A(new BallotNumber(ballot.sequenceNum(), address()), i, new PaxosRequest(null));
                        for (Address server : servers) {
                            if (server != address()) {
                                send(p2a, server);
                            }
                        }
                        slotsToAcceptors.put(i, address());
                        proposals.put(i, new BallotValuePair(new BallotNumber(ballot.sequenceNum(), address()), p2a.value()));
                        accepted.put(i, new BallotValuePair(new BallotNumber(ballot.sequenceNum(), address()), p2a.value()));
                    }
                }
            }
        }
    }

    private int garbageCollectUntil() {
        if (replicasLastExecuted.keySet().size() != servers.length) {
            return 0;
        }
        int min = Integer.MAX_VALUE;
        for (int i : replicasLastExecuted.values()) {
            min = Math.min(i, min);
        }
        return min;
    }

    @Data
    static class BallotValuePair implements Comparable, Serializable{
        private final BallotNumber ballotNumber;
        private final PaxosRequest value;


        @Override
        public int compareTo(Object o) {
            BallotValuePair bvp = (BallotValuePair) o;
            return ballotNumber().compareTo(bvp.ballotNumber());
        }
    }

    @Data
    static class BallotNumber implements Comparable, Serializable {
        private int sequenceNum;
        private Address address;

        public BallotNumber(int sequenceNum, Address address) {
            this.sequenceNum = sequenceNum;
            this.address = address;
        }

        private void increaseBallot() {
            sequenceNum++;
        }

        @Override
        public int compareTo(Object o) {
            // Check o is ballotNumber
            BallotNumber b = (BallotNumber) o;
            if (sequenceNum > b.sequenceNum) {
                return 1;
            } else if (sequenceNum == b.sequenceNum){
                return address.compareTo(b.address);
            } else {
                return -1;
            }
        }
    }


}

