# http://pymotw.com/2/SocketServer/
from functools import total_ordering
from message import *
import socketserver
import pickle
import socket
from collections import defaultdict

# Paxos servers
class Paxos(socketserver.TCPServer):
    def __init__(self, address, servers) -> None:
        # address is a tuple of (ip, port)
        self.address = address                                      # Our address (ip, port)
        self.servers = servers                                      # All servers addresses

        # Sets up other variables
        self.is_leader = False                                      # Are we the leader?
        self.ballot = BallotNumber(1, self.address)                 # Our ballot number
        self.highest_ballot_seen = BallotNumber(0, self.address)    # Highest ballot number seen
        self.proposals = dict()                                     # Used by leader to send P2A to acceptors
        self.accepted = dict()                                      # Used by acceptors in P2A and also P1B
        self.log = dict()                                           # The log
        self.leader_recent_ping = False

        # for proposal phase
        self.slot_in = 1                                            # First non-executed proposal slot (+1 last executed slot)
        self.slot_out = 1                                           # First empty slot (+1 last proposed slot)

        # Multimap equivalent in python is defaultdict: https://stackoverflow.com/questions/1731971/is-there-a-multimap-implementation-in-python
        self.slot_to_acceptors = defaultdict(set)                   # Used by leader to decide majority for each slot after P2B

        # for leader election
        self.voters = set()                                         # Yes votes for leader election
        self.p1b_replies = dict()                                   # Accepted values for each slot by acceptors that voted yes
        # ...

        # Default leader during setup
        if self.address == self.servers[0]:
            self.is_leader = True
        
        print("Finished init paxos")
        print("servers: " + str(self.servers))
        print("address: " + str(self.address))
        print("is_leader: " + str(self.is_leader))

        socketserver.TCPServer.__init__(self, address, PaxosHandler)


class PaxosHandler(socketserver.BaseRequestHandler):
    """
    The request handler class for our server.

    It is instantiated once per connection to the server, and must
    override the handle() method to implement communication to the
    client.

    This will receive lock() / unlock() command from client
    Handler for proposals and leader election will be a different class (I think, probably using different port?)
    """

    def handle(self):
        # Use self.server.arg to get servers fields
        # Note that we guarantee communication client to server is exactly once,
        # no need to worry about duplicate request and proposing two slot. 
        data = self.request.recv(1024).strip() # DEBUG LOG
        print("{} wrote:".format(self.client_address)) # DEBUG LOG
        print(data) # DEBUG LOG
        
        # test deserialize
        message = pickle.loads(data)

        print(message, "\n") # debug
        if isinstance(message, PaxosRequest):
            self.handlePaxosRequest(message, self.client_address)
            print("got paxos request") # debug

        elif isinstance(message, P1A):
            print("got p1a")    # debug
            self.handleP1A(message, self.client_address)

        elif isinstance(message, P1B):
            print("got p1b")    # debug
            self.handleP1B(message, self.client_address)

        elif isinstance(message, P2A):
            print("got p2a")    # debug
            self.handleP2A(message, self.client_address)

        elif isinstance(message, P2B):
            print("got p2b")    # debug
            self.handleP2B(message, self.client_address)
        else:
            # TODO: add cases for heartbeats
            # prob just ignore message
            print("unrecognized message")   # debug
            
    def handlePaxosRequest(self, paxos_req, sender):
        print(self.server.is_leader)
        if not self.server.is_leader:
            return
        print("Handling paxos request")
        
        # TODO: add lockManager app here, translate below java code (No need to check already executed, etc, for AMO)
        #if (app.alreadyExecuted(m.value())) {
            #AMOResult r = app.execute(m.value());
            #if (r != null) {
                #send(new PaxosReply(r), r.clientAddress());
            #}
            #return;
        #}
        #if (requestInProcess.contains(m)) {
            #// Used to make sure shat we dont use two log slot for the same
            #// request
            #return;
        #}

        bvp = BallotValuePair(BallotNumber(self.server.ballot.seq_num, self.server.address), paxos_req)
        self.server.proposals[self.server.slot_in] = bvp
        self.server.accepted[self.server.slot_out] = bvp

        p2a = P2A(self.server.address, BallotNumber(self.server.ballot.seq_num, self.server.address), self.server.slot_in, paxos_req)

        # increase slot_in
        self.server.slot_in += 1

        # P2A
        for acceptor in self.server.servers:
            if acceptor != self.server.address:
                send_msg(p2a, acceptor)

        # accept our own proposal
        self.server.slot_to_acceptors[p2a.slot_num].add(self.server.address)

        # TODO: translate java code
        #requestInProcess.add(p2a.value()); No need, this is for AMO

        #set(new P2ATimer(p2a), P2ATimer.P2A_RETRY_MILLIS);
        if len(self.server.slot_to_acceptors[p2a.slot_num]) > (len(self.server.servers)/2):
            #Majority accepted, can put into log
            self.server.log[p2a.slot_num] = bvp
            self.request.sendall(pickle.dumps(PaxosResult((HOST, 12345), True)))
            # TODO: Executelog() function
            # executeLog();

        # Testing code to return the PaxosResult
        # self.request.sendall(pickle.dumps(PaxosResult((HOST, 12345), True)))

    def handleP1A(self, p1a, sender):
        if self.server.highest_ballot_seen < p1a.ballot_num:
            self.server.highest_ballot_seen = p1a.ballot_num

            # If we are leader, make this a follower (since the one sending P1A thinks they are leader)
            if self.server.is_leader:
                self.server.is_leader = False
            
            p1b = P1B(self.server.address, self.server.highest_ballot_seen, self.server.accepted)
            send_msg(p1b, sender)
            #TODO: Set timer for heartbeatcheck
            self.leader_recent_ping = True

    def handleP1B(self, p1b, sender):
        if self.server.is_leader:
            return
        # If receive majority of response from acceptors with its ballot, becomes leader
        if self.server.ballot == p1b.accepted_ballot and self.server.ballot == self.server.highest_ballot_seen:
            self.voters.add(sender)
            for slot in p1b.accepted:
                new_bvp = p1b[slot]
                cur_bvp = self.server.p1b_replies[slot]
                if cur_bvp is None:
                    self.server.p1b_replies[slot] = new_bvp
                else:
                    if new_bvp.ballot_num >= cur_bvp.ballot_num:
                        self.server.p1b_replise[slot] = new_bvp
            if len(self.server.voters) > (len(self.server.servers)/2 - 1):
                # This server is elected as leader
                self.server.is_leader = True
                # Must update its state with accepted values from acceptors
                for slot in self.server.p1b_replies:
                    pass
                    # TODO Translate java code
                    # if (status(slot) != PaxosLogSlotStatus.CHOSEN) {
                    #     proposals.put(slot, new BallotValuePair(new BallotNumber(ballot.sequenceNum(), address()), p1bReplies.get(slot).value()));
                    #     slotIn = Math.max(slotIn, slot + 1);
                    #     BallotValuePair bvp = proposals.get(slot);

                    #     P2A p2a = new P2A(bvp.ballotNumber(), slot, bvp.value());
                    #     for (Address acceptor : servers) {
                    #         if (!acceptor.equals(address())) {
                    #             send(p2a, acceptor);
                    #         }
                    #     }
                    #     accepted.put(slot, new BallotValuePair(new BallotNumber(ballot.sequenceNum(), address()), p1bReplies.get(slot).value()));
                    #     // Accept our own proposal
                    #     slotsToAcceptors.put(p2a.slotNum(), address());

                    #     requestInProcess.add(p2a.value());
                    #     set(new P2ATimer(p2a), P2ATimer.P2A_RETRY_MILLIS);
                    # }
    
    def handleP2A(self, p2a, sender):
        if self.server.highest_ballot_seen != p2a.ballot:
            # It's not the leader, drop it
            return
        # TODO: Not sure if needed but translate Java code
        # if (status(p2a.slotNum()) == PaxosLogSlotStatus.CLEARED) {
        #     return;
        # }
        if p2a.slot_num in self.server.accepted:
            bvp = self.server.accepted[p2a.slot_num]
            if bvp.ballot <= p2a.ballot:
                # p2a ballot is higher or equal
                bvp = BallotValuePair(p2a.ballot, p2a.value)
                self.server.accepted[p2a.slot_num, bvp]
            else:
                # Don't do anything
                return
        else:
            # Have not accepted anything, then accept it
            bvp = BallotValuePair(p2a.ballot, p2a.value)
            self.server.accepted[p2a.slot_num, bvp]
        
        self.server.slot_in = max(self.server.slot_in, p2a.slot_num + 1)

        p2b = P2B(p2a.ballot, p2a.slot_num)
        send_msg(p2b, sender)

    def handleP2B(self, p2b, sender):
        if not self.server.is_leader:
            # Not leader, drop message
            return

        # TODO: Not sure if needed but translate Java code
        # if (status(p2b.slotNum()) == PaxosLogSlotStatus.CLEARED) {
        #     return;
        # }

        # check if it is still consistent with our proposal
        bvp = self.server.proposals(p2b.slot_num)
        if bvp.ballot_num != p2b.ballot:
            # No longer in proposal
            return

        # Keep track of who have accepted
        self.server.slot_to_acceptors[p2b.slot_num].add(sender)
        if len(self.server.slot_to_acceptors[p2b.slot_num]) > (len(self.server.servers)/2):
            # Majority accepted, can put into log
            self.server.log[p2b.slot_num] = bvp
            # TODO: Create executeLog() and uncomment below
            # executeLog()


@total_ordering
class BallotNumber:
    def __init__(self, seq_num, addr) -> None:
        self.seq_num = seq_num # Sequence number of the proposal
        self.addr = addr # Address of the server proposing

    def increaseBallot(self):
        self.seq_num += 1

    def _is_valid_operand(self, other):
        return (hasattr(other, "seq_num") and
                hasattr(other, "addr"))

    # Only need to make lt and eq, total_ordering will do the rest
    def __lt__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        if self.seq_num < other.seq_num:
            return True
        elif self.seq_num > other.seq_num:
            return False
        else:
            # Same seq_num, use address for break
            return self.addr[1] < other.addr[1]
        
    def __eq__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        return ((self.seq_num, self.addr) == (other.seq_num, other.addr))

    def __str__(self) -> str:
        return "BallotNumber(seq_num: " + str(self.seq_num) + ", addr: " + str(self.addr) + ")"
        
@total_ordering
class BallotValuePair:
    def __init__(self, ballot_num, value) -> None:
        self.ballot_num = ballot_num # ballot number
        self.value = value # paxos request

    def _is_valid_operand(self, other):
        return (hasattr(other, "ballot_num") and
                hasattr(other, "value"))

    # Only need to make lt and eq, total_ordering will do the rest
    def __lt__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        
        return self.ballot_num < other.ballot_num
        
    def __eq__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        return self.ballot_num == other.ballot_num

    def __str__(self) -> str:
        return "BallotValuePair(ballot_num: " + str(self.ballot_num) + ", value: " + str(self.value) + ")"

# Serialize obj, and send the message.
# This function will not wait for reply (communication between paxos nodes)
def send_msg(obj, dest_addr):
    HOST, PORT = dest_addr

    data = pickle.dumps(obj)
    print(obj, "\n")
    print(data, "\n")
    de_paxos_req = pickle.loads(data)
    print(de_paxos_req, "\n")
    #data = " ".join(sys.argv[1:])
    #print('data = %s' %data)
    # create a TCP socket
    # TODO: Between paxos dont use TCP? Use UDP instead? TCP will make it does not drop any packet right?
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    try:
        # connect to server 
        sock.connect((HOST, PORT))

        # send data
        #sock.sendall(bytes(data + "\n"))
        sock.sendall(data)
    finally:
        # shut down
        sock.close()

if __name__ == "__main__":
    HOST, PORT = "localhost", 9999
    # addresses = ((HOST, PORT), (HOST, PORT + 1), (HOST, PORT + 2))
    addresses = ((HOST, PORT),)
    # Create the server, binding to localhost on port 9999
    server = Paxos((HOST, PORT), addresses)
    # Activate the server; this will keep running until you
    # interrupt the program with Ctrl-C
    server.serve_forever()
